// To parse this data:
//
//   import { Convert, Payload, Envelope, PayloadT, EnvelopeT, PayloadT1 } from "./file";
//
//   const payload = Convert.toPayload(json);
//   const envelope = Convert.toEnvelope(json);
//   const payloadT = Convert.toPayloadT(json);
//   const t = Convert.toT(json);
//   const envelopeT = Convert.toEnvelopeT(json);
//   const payloadT1 = Convert.toPayloadT1(json);
//   const t1 = Convert.toT1(json);
//
// These functions will throw an error if the JSON doesn't
// match the expected interface, even if the JSON is valid.

export interface PayloadT {
    data: { [key: string]: any };
    kind: string;
}

export interface EnvelopeT {
    data: PayloadT1;
    dst:  string[];
    id:   string;
    src:  string;
    t:    number;
    ttl:  number;
    v:    V;
}

export interface PayloadT1 {
    data: { [key: string]: any };
    kind: string;
}

export enum V {
    A = "A",
}

// Converts JSON strings to/from your types
// and asserts the results of JSON.parse at runtime
export class Convert {
    public static toPayload(json: string): PayloadT {
        return cast(JSON.parse(json), r("PayloadT"));
    }

    public static payloadToJson(value: PayloadT): string {
        return JSON.stringify(uncast(value, r("PayloadT")), null, 2);
    }

    public static toEnvelope(json: string): EnvelopeT {
        return cast(JSON.parse(json), r("EnvelopeT"));
    }

    public static envelopeToJson(value: EnvelopeT): string {
        return JSON.stringify(uncast(value, r("EnvelopeT")), null, 2);
    }

    public static toPayloadT(json: string): PayloadT {
        return cast(JSON.parse(json), r("PayloadT"));
    }

    public static payloadTToJson(value: PayloadT): string {
        return JSON.stringify(uncast(value, r("PayloadT")), null, 2);
    }

    public static toT(json: string): { [key: string]: any } {
        return cast(JSON.parse(json), m("any"));
    }

    public static tToJson(value: { [key: string]: any }): string {
        return JSON.stringify(uncast(value, m("any")), null, 2);
    }

    public static toEnvelopeT(json: string): EnvelopeT {
        return cast(JSON.parse(json), r("EnvelopeT"));
    }

    public static envelopeTToJson(value: EnvelopeT): string {
        return JSON.stringify(uncast(value, r("EnvelopeT")), null, 2);
    }

    public static toPayloadT1(json: string): PayloadT1 {
        return cast(JSON.parse(json), r("PayloadT1"));
    }

    public static payloadT1ToJson(value: PayloadT1): string {
        return JSON.stringify(uncast(value, r("PayloadT1")), null, 2);
    }

    public static toT1(json: string): { [key: string]: any } {
        return cast(JSON.parse(json), m("any"));
    }

    public static t1ToJson(value: { [key: string]: any }): string {
        return JSON.stringify(uncast(value, m("any")), null, 2);
    }
}

function invalidValue(typ: any, val: any, key: any = ''): never {
    if (key) {
        throw Error(`Invalid value for key "${key}". Expected type ${JSON.stringify(typ)} but got ${JSON.stringify(val)}`);
    }
    throw Error(`Invalid value ${JSON.stringify(val)} for type ${JSON.stringify(typ)}`, );
}

function jsonToJSProps(typ: any): any {
    if (typ.jsonToJS === undefined) {
        const map: any = {};
        typ.props.forEach((p: any) => map[p.json] = { key: p.js, typ: p.typ });
        typ.jsonToJS = map;
    }
    return typ.jsonToJS;
}

function jsToJSONProps(typ: any): any {
    if (typ.jsToJSON === undefined) {
        const map: any = {};
        typ.props.forEach((p: any) => map[p.js] = { key: p.json, typ: p.typ });
        typ.jsToJSON = map;
    }
    return typ.jsToJSON;
}

function transform(val: any, typ: any, getProps: any, key: any = ''): any {
    function transformPrimitive(typ: string, val: any): any {
        if (typeof typ === typeof val) return val;
        return invalidValue(typ, val, key);
    }

    function transformUnion(typs: any[], val: any): any {
        // val must validate against one typ in typs
        const l = typs.length;
        for (let i = 0; i < l; i++) {
            const typ = typs[i];
            try {
                return transform(val, typ, getProps);
            } catch (_) {}
        }
        return invalidValue(typs, val);
    }

    function transformEnum(cases: string[], val: any): any {
        if (cases.indexOf(val) !== -1) return val;
        return invalidValue(cases, val);
    }

    function transformArray(typ: any, val: any): any {
        // val must be an array with no invalid elements
        if (!Array.isArray(val)) return invalidValue("array", val);
        return val.map(el => transform(el, typ, getProps));
    }

    function transformDate(val: any): any {
        if (val === null) {
            return null;
        }
        const d = new Date(val);
        if (isNaN(d.valueOf())) {
            return invalidValue("Date", val);
        }
        return d;
    }

    function transformObject(props: { [k: string]: any }, additional: any, val: any): any {
        if (val === null || typeof val !== "object" || Array.isArray(val)) {
            return invalidValue("object", val);
        }
        const result: any = {};
        Object.getOwnPropertyNames(props).forEach(key => {
            const prop = props[key];
            const v = Object.prototype.hasOwnProperty.call(val, key) ? val[key] : undefined;
            result[prop.key] = transform(v, prop.typ, getProps, prop.key);
        });
        Object.getOwnPropertyNames(val).forEach(key => {
            if (!Object.prototype.hasOwnProperty.call(props, key)) {
                result[key] = transform(val[key], additional, getProps, key);
            }
        });
        return result;
    }

    if (typ === "any") return val;
    if (typ === null) {
        if (val === null) return val;
        return invalidValue(typ, val);
    }
    if (typ === false) return invalidValue(typ, val);
    while (typeof typ === "object" && typ.ref !== undefined) {
        typ = typeMap[typ.ref];
    }
    if (Array.isArray(typ)) return transformEnum(typ, val);
    if (typeof typ === "object") {
        return typ.hasOwnProperty("unionMembers") ? transformUnion(typ.unionMembers, val)
            : typ.hasOwnProperty("arrayItems")    ? transformArray(typ.arrayItems, val)
            : typ.hasOwnProperty("props")         ? transformObject(getProps(typ), typ.additional, val)
            : invalidValue(typ, val);
    }
    // Numbers can be parsed by Date but shouldn't be.
    if (typ === Date && typeof val !== "number") return transformDate(val);
    return transformPrimitive(typ, val);
}

function cast<T>(val: any, typ: any): T {
    return transform(val, typ, jsonToJSProps);
}

function uncast<T>(val: T, typ: any): any {
    return transform(val, typ, jsToJSONProps);
}

function a(typ: any) {
    return { arrayItems: typ };
}

function u(...typs: any[]) {
    return { unionMembers: typs };
}

function o(props: any[], additional: any) {
    return { props, additional };
}

function m(additional: any) {
    return { props: [], additional };
}

function r(name: string) {
    return { ref: name };
}

const typeMap: any = {
    "PayloadT": o([
        { json: "data", js: "data", typ: m("any") },
        { json: "kind", js: "kind", typ: "" },
    ], "any"),
    "EnvelopeT": o([
        { json: "data", js: "data", typ: r("PayloadT1") },
        { json: "dst", js: "dst", typ: a("") },
        { json: "id", js: "id", typ: "" },
        { json: "src", js: "src", typ: "" },
        { json: "t", js: "t", typ: 3.14 },
        { json: "ttl", js: "ttl", typ: 3.14 },
        { json: "v", js: "v", typ: r("V") },
    ], "any"),
    "PayloadT1": o([
        { json: "data", js: "data", typ: m("any") },
        { json: "kind", js: "kind", typ: "" },
    ], "any"),
    "V": [
        "A",
    ],
};
